from .hook import Hook
from ..utils import convert_to_api_class_name


class HooksManager:
	def __init__(self, agent):
		self._agent = agent
		self._hooks = {}
		self._inc_id = 1

	def __repr__(self):
		return str(self._hooks)

	def __getitem__(self, key):
		return self._hooks[key]

	def disable(self, hook_id):
		self._agent.api.disable_hook(hook_id)
		self._hooks[hook_id].enabled = False

	def enable(self, hook_id):
		self._agent.api.enable_hook(hook_id)
		self._hooks[hook_id].enabled = True

	def call(self, hook_id, params, orig_retval):
		return self._hooks[hook_id].hook_impl(params, orig_retval)

	def add(self, class_name, method_name, param_types=None, hook_impl=None, get_original_retval=False,
	        show_output=False):
		if show_output:
			self._agent.output.start()

		hook_id = self._inc_id

		class_name = convert_to_api_class_name(class_name)
		if param_types:
			param_types = [convert_to_api_class_name(pt) for pt in param_types]

		if hook_impl is None:
			def default_hook_impl(params, original_retval):
				self._agent.__print__(params)
				return original_retval

			get_original_retval = True
			hook_impl = default_hook_impl

		def wrap_hook_func_with_args_convertion(hook_impl_func):
			def hook_impl_wrapper(args_json_list, original_retval):
				args = list(map(lambda arg_json: self._agent._value_data_to_python_value(arg_json), args_json_list))
				orig_retval_python = self._agent._value_data_to_python_value(original_retval)
				val = hook_impl_func(args, orig_retval_python)
				return val

			return hook_impl_wrapper

		hook_impl = wrap_hook_func_with_args_convertion(hook_impl)

		hook = Hook(self, hook_id, class_name, method_name, param_types, hook_impl, show_output)
		self._hooks[hook_id] = hook
		self._inc_id += 1

		self._agent.api.hook_method(hook_id, class_name, method_name, param_types, get_original_retval)

		return hook



"""
Represents a connected Android Device
"""

import re
from time import sleep

from .android_manifest import AndroidManifest
from .app import App
from .consts import *
from .files import FilesOperations
from .log import *
from .storage import Storage
from .utils import *


class Device:

	def __get_all_apps__(self):
		res = self.shell("pm list packages -f")
		res = res.strip().split('\n')
		res = list(map(lambda x: x[len('package:'):], res))
		apps = []
		for p in res:
			eq_idx = p.rfind('=')
			remote_apk_path = p[:eq_idx]
			package_name = p[eq_idx + 1:]

			app = App(self, package_name, remote_apk_path)

			apps.append(app)

		return apps

	def __log__(self, s, level=Log.INFO, extra_info=None):
		to_print = Fore.GREEN + '[' + self.id + '] ' + Style.RESET_ALL + s

		if extra_info:
			to_print += " [" + Style.BRIGHT + Fore.MAGENTA + str(
				extra_info) + Style.RESET_ALL + "]"

		log(to_print, level=level, do_style=False)

	def __warn__(self, s, extra_info=None):
		self.__log__(Fore.YELLOW + s, Log.WARN, extra_info)

	def __dlog__(self, s, extra_info=None):
		self.__log__(Fore.BLUE + Style.BRIGHT + s, Log.DEBUG, extra_info)

	def __error__(self, s):
		self.__log__(Fore.RED + Style.BRIGHT + s, Log.ERROR)

	def __success__(self, s):
		self.__log__(Fore.GREEN + Style.BRIGHT + s, Log.INFO)

	def __adb_prefix__(self):
		return '"{0}" -s {1}'.format(self.adb_path, self.id)

	def __init__(self, serial, adb_path='adb'):
		self.id = serial
		self.adb_path = adb_path
		self.files = FilesOperations(self)

	def __repr__(self):
		return self.id

	def summary(self):
		"""
		Prints some details about the device.
		"""

		def aux_print(name, value):
			pad = 15
			self.__log__(Fore.CYAN + name.ljust(pad) + Style.RESET_ALL + str(value))

		self.__log__("Device summary:")
		aux_print("	Arch", self.get_arch())
		aux_print("	Android ver.", self.get_android_version())
		aux_print("	API level", self.get_api_level())
		aux_print("	Is rooted", self.is_rooted())

	def shell(self, cmd, wait=True, as_root=False):
		"""
		Run shell command on the device.

		:param cmd: The cmd to run.
		:type cmd: str
		:param wait: Should wait for the execution to finish
		:type wait: bool
		:param as_root: Run the command as root (supported only on rooted device with `su` binary)
		:type as_root: bool
		:return: The output if wait=True, the Popen object otherwise.
		:rtype: str
		"""
		cmd = '{0} shell {1} "{2}"'.format(self.__adb_prefix__(), "su root" if as_root else "", cmd)

		if wait:
			return check_output(cmd)

		return run_cmd(cmd)

	def get_apps(self, name_substr=None):
		"""
		Get a list of apps installed on the device.
		:param name_substr: Specify to return apps which their name contains this string
		:type name_substr: str
		:return: A list of apps
		:rtype: List of :py:class:`~marrs.app.App`s
		"""
		apps = self.__get_all_apps__()

		if name_substr:
			apps = list(filter(lambda app: name_substr in app.name, apps))

		return apps

	def get_app_version(self, app_name):
		"""
		Get application installed version.

		:param app_name: The full name of the app
		:type app_name: str
		:return: The version of the given app
		:rtype: str
		"""
		if not self.get_app(app_name):
			return None

		res = self.shell('dumpsys package {0} | grep versionName'.format(app_name))
		p = re.compile(r'versionName=([\.\d]+)')
		return p.findall(res)[0]

	def install_app(self, apk_path, force_install=False):
		"""
		Install an APK file on the device.

		:param apk_path: The path to the APK file
		:type apk_path: str
		:param force_install: If True, will install the APK if the app with the same version is already installed on the device
		:type force_install: bool
		:return: The :py:class:`~marrs.app.App` object of the installed app
		:rtype: :py:class:`~marrs.app.App`
		"""
		self.__log__("Installing app from apk ({0})".format(apk_path))

		manifest = AndroidManifest.from_apk(apk_path)
		app_name = manifest.get_package_name()

		self.__log__("	Installing app", extra_info=app_name)

		app = self.get_app(app_name)

		if app:
			current_version = app.get_version()
			version_to_install = manifest.get_package_version()

			if not force_install and current_version == version_to_install:
				self.__warn__(
					"	The package '{0}' with the same version ({1}) is already installed on the device".format(
						app_name, current_version))
				return app

			self.__warn__("	Package already installed on device, uninstalling version {0}".format(current_version))

			app.uninstall()

		check_output('{0} install "{1}"'.format(self.__adb_prefix__(), apk_path))

		return self.get_app(app_name)

	def is_app_installed(self, app_name):
		"""
		Check if app is installed on the device.

		:param app_name: The name of the application
		:type app_name: str
		:return: True if the app is installed on the device, False otherwise.
		:rtype: bool
		"""
		return self.get_app(app_name) is not None

	def uninstall_app(self, app_name):
		"""
		Uninstall application from the device.

		:param app_name: The application's name to uninstall.
		:type app_name: str
		"""
		self.__log__("Uninstalling app", extra_info=app_name)

		if not self.is_app_installed(app_name):
			self.__warn__("  App is not installed")
			return

		check_output("{0} uninstall {1}".format(self.__adb_prefix__(), app_name))

	def get_app(self, app_name):
		"""
		Get app by name.

		:param app_name: The name of the app
		:type app_name: str
		:return: :py:class:`~marrs.app.App` object if app with the given name is installed on the device, None otherwise.
		:rtype: :py:class:`~marrs.app.App`
		"""
		apps = self.__get_all_apps__()

		for app in apps:
			if app.name == app_name:
				return app

		return None

	def is_app_running(self, app_name):
		"""
		Check if app is running.

		:param app_name: The name of the app
		:type app_name: str
		:return: True if app is running, False otherwise.
		:rtype: bool
		"""
		try:
			self.shell("pidof {0}".format(app_name))
			return True
		except:
			return False

	def is_rooted(self):
		"""
		Check if device is rooted (by trying to run 'su root').

		:return: True if device is rooted, False otherwise.
		:rtype: bool
		"""
		return self.shell('su root \\"echo\\" > /dev/null 2>&1; echo $?') == '0'

	def get_android_version(self):
		"""
		Returns the android version of the device.

		:return: Android version of the device
		:rtype: str
		"""
		return self.shell("getprop ro.build.version.release")

	def get_api_level(self):
		"""
		Returns the API level of the device.

		:return: API level of the device
		:rtype: str
		"""
		return self.shell("getprop ro.build.version.sdk")

	def get_arch(self):
		"""
		Returns the architechture of the device (arm64 / arm / x86_64 / x86)

		:return: The architecture of the device
		:rtype: str
		"""
		res = self.shell("getprop ro.product.cpu.abi").lower()
		archs = ['arm64', 'arm', 'x86_64', 'x86']
		for arch in archs:
			if arch in res:
				return arch

		raise Exception('Failed identify arch : `{0}`'.format(res))

	def is_frida_server_running(self, frida_process_name='frida-server'):
		"""
		Check if frida server is running on the device.

		:param frida_process_name: The name of the frida server process
		:type frida_process_name: str
		:return: True if frida server is running, False otherwise.
		:rtype: bool
		"""
		return frida_process_name in self.shell("ps")

	def kill_frida_server(self):
		"""
		Kills frida server's procces.

		:raises: Exception: If failed to kill the process
		"""

		self.__log__("Killing frida server process")

		if not self.is_frida_server_running():
			self.__warn__("Frida server is not running on device")
			return

		self.shell("pkill frida-server", as_root=True)

		sleep(1)
		self.shell("echo")

		if self.is_frida_server_running():
			raise Exception("Failed to kill frida server")

	def delete_frida_server(self):
		"""
		Deletes all the files which their path mathches the following pattern: '/data/local/tmp/frida-server-\*'
		"""
		self.shell("rm /data/local/tmp/frida-server-* 1> /dev/null 2>&1")

	def __upload_frida_server__(self, version=FRIDA_SERVER_VERSION, arch=None):
		if not arch:
			arch = self.get_arch()

		extra_info = "arch={arch}, version={version}".format(
			arch=arch, version=version)

		self.__log__("Uploading frida server to device", extra_info=extra_info)

		frida_server_filename = FRIDA_SERVER_FILENAME.format(
			version=version, arch=arch)
		self.__dlog__("frida_server_filename = " + frida_server_filename)

		if Storage.contains(frida_server_filename):
			self.__log__(
				"Frida server exists in local storage", extra_info=extra_info)

			frida_server_local_filepath = Storage.get(frida_server_filename)

		else:
			self.__dlog__(
				"Frida server doesn't exist in local storage", extra_info=extra_info)
			self.__log__(
				"Trying to download it from the internet", extra_info=extra_info)

			frida_server_url = FRIDA_SERVER_DOWNLOAD_URL.format(
				version=version, filename=frida_server_filename)
			self.__dlog__("frida_server_url = " + frida_server_url)

			filename = get_filename_from_url(frida_server_url)
			dest_file_path = path.join(Storage.create_temp_dir(), filename)
			frida_server_file_path = download_from_url(frida_server_url, dest_file_path)

			if not frida_server_file_path:
				raise Exception("Failed download frida server from the internet")

			self.__dlog__("frida_server_file_path = " + frida_server_file_path)

			Storage.save(frida_server_file_path)

			frida_server_local_filepath = extract_xz_file(frida_server_file_path)

		self.__dlog__("frida_server_local_filepath = " + frida_server_local_filepath)

		frida_server_remote_path = FRIDA_SERVER_REMOTE_PATH.format(
			version=version, arch=arch)
		self.__dlog__("frida_server_remote_path = " + frida_server_remote_path)
		self.files.push(frida_server_local_filepath, frida_server_remote_path)

	def run_frida_server(self, frida_server_version=FRIDA_SERVER_VERSION):
		"""
		Run frida server of the given version on the device.
		First it will check if the binary is already exists on the device. If so, will run it.
		If it doesn't exist, it will check if the binary exists in the package's store. If so, it will take it from there.
		Otherwise, it will try to download the binary from the internet, upload it and run it on the device.

		:param frida_server_version: the frida server's version to run on the device
		:type frida_server_version: str
		:raise Exception: If the device is not rooted or if failed to run frida server on the device.
		"""
		self.__log__("Running frida server on device")

		if self.is_frida_server_running():
			self.__warn__("	Frida server is already running on device")
			return

		if not self.is_rooted():
			raise Exception("Device must be rooted in order to run frida server")
			return

		arch = self.get_arch()
		extra_info = "arch={arch}, version={version}".format(
			arch=arch, version=frida_server_version)

		frida_server_remote_path = FRIDA_SERVER_REMOTE_PATH.format(
			version=frida_server_version, arch=arch)

		if self.files.exists(frida_server_remote_path):
			self.__log__(
				"	Frida server file is already exists on device",
				extra_info=extra_info)
		else:
			self.__warn__("	Frida server file doesn't exist on device")
			self.__upload_frida_server__(frida_server_version, arch)

		self.shell("chmod +x {0}".format(frida_server_remote_path), as_root=True)
		self.shell(
			"nohup .{0} &".format(frida_server_remote_path),
			wait=False,
			as_root=True)

		sleep(1)

		if not self.is_frida_server_running():
			raise Exception("Failed run frida server on the device")

		self.__success__("	Frida server is now running on device")

	def start_app(self, app_name, main_activity=None):
		"""
		Start application on the device.
		Will do nothing if app is already running.

		:param app_name: The name of the app to start.
		:type app_name: str
		:param main_activity: the main activity of the application to start. If None, it will extract it from the app's apk that exists on the device.
		:type main_activity: str
		:raise Exception: If the device is not rooted or if failed to run frida server on the device.
		"""
		self.__log__("Starting app", extra_info=app_name)

		if self.is_app_running(app_name):
			self.__warn__("	App is already running on device")
			return

		if not main_activity:
			app = self.get_app(app_name)
			main_activity = app.get_main_activity()

		self.__log__("	Main activity is '{0}'".format(main_activity))

		self.shell("am start -n {0}/{1}".format(app_name, main_activity))

		sleep(1)

	def force_stop_app(self, app_name):
		"""
		Force stop application.

		:param app_name: The name of the app to force stop.
		:type app_name: str
		"""
		self.__log__("Force stop app", extra_info=app_name)
		self.shell("am force-stop {0}".format(app_name))

	def clear_app_data(self, app_name):
		"""
		Clear application data.

		:param app_name: The name of the app to clear data for.
		:type app_name: str
		"""
		self.__log__("Clearing app data", extra_info=app_name)

		self.shell("pm clear {0}".format(app_name))



import codecs
import lzma
import pathlib
import shutil
import subprocess
import urllib.request
from os import path, makedirs
from tempfile import _get_candidate_names
from urllib.parse import urlparse
from zipfile import ZipFile


def copy_file_with_dir_creation(src_filepath, dest_filepath):
	makedirs(get_dir_path(dest_filepath), exist_ok=True)
	shutil.copyfile(src_filepath, dest_filepath)


def read_file(file_path):
	with codecs.open(file_path, 'r', 'utf-8') as f:
		return f.read()


def run_cmd(cmd):
	subprocess.Popen(
		cmd,
		shell=True,
		encoding='utf8',
		stdout=subprocess.PIPE,
		stderr=subprocess.PIPE)


def delete_folder(folder_path):
	shutil.rmtree(folder_path)


def check_output(cmd):
	proc = subprocess.run(
		cmd,
		shell=True,
		check=False,
		encoding='utf8',
		stdout=subprocess.PIPE,
		stderr=subprocess.PIPE)

	if proc.returncode != 0:
		raise Exception(
			"Run command `{0}` failed, stderr=`{1}`, stdout=`{2}`, returncode=`{3}`"
				.format(cmd, proc.stderr, proc.stdout, proc.returncode))

	return proc.stdout.strip()


def get_filename_from_url(url):
	return path.basename(urlparse(url).path)


def invert_dict(d):
	return {v: k for k, v in d.items()}


def download_from_url(file_url, dest_file_path):
	with urllib.request.urlopen(file_url) as response, open(dest_file_path,
	                                                        'wb') as out_file:
		shutil.copyfileobj(response, out_file)
	return dest_file_path


def get_temp_name():
	return next(_get_candidate_names())


def get_dir_path(file_path):
	return pathlib.Path(file_path).parent.resolve()


def get_filename(file_path, with_extension=True):
	filename = path.basename(file_path)

	if not with_extension:
		filename = remove_extension(filename)

	return filename


def remove_extension(filepath):
	return path.splitext(filepath)[0]


def extract_xz_file(xzfile_path, dest_file_path=None):
	if not dest_file_path:
		dest_file_path = remove_extension(xzfile_path)

	with lzma.open(xzfile_path) as f, open(dest_file_path, 'wb') as fout:
		file_content = f.read()
		fout.write(file_content)

	return dest_file_path


def extract_zip_file(zipfile_path, dest_dir_path):
	with ZipFile(zipfile_path, 'r') as zipObj:
		# Extract all the contents of zip file in current directory
		zipObj.extractall(dest_dir_path)

	return dest_dir_path



from .instances import Array, Instance


class InstanceFactory:
	@staticmethod
	def create(agent, instance_data):
		if instance_data.get('is_array'):
			return Array(instance_data, agent)

		return Instance(instance_data, agent)



import tempfile
from os import path

from androguard.core.bytecodes.axml import AXMLPrinter
from lxml import etree

from .utils import extract_zip_file


class AndroidManifest:
	ANDROID_NAMESPACE = '{http://schemas.android.com/apk/res/android}'

	@staticmethod
	def from_apk(apk_path):
		with tempfile.TemporaryDirectory() as tmp_dir_path:
			extract_zip_file(apk_path, tmp_dir_path)
			manifest_path = path.join(tmp_dir_path, 'AndroidManifest.xml')
			manifest = AndroidManifest(manifest_path)
			return manifest

	def __init__(self, file_path):
		with open(file_path, "rb") as fp:
			axml = AXMLPrinter(fp.read())
			# Get the lxml.etree.Element from the AXMLPrinter:
			self.root = axml.get_xml_obj()

	def __str__(self):
		s = etree.tostring(self.root).decode('utf8')
		return s

	def get_package_name(self):
		return self.root.attrib['package']

	def get_package_version(self):
		return self.root.attrib[AndroidManifest.ANDROID_NAMESPACE + 'versionName']

	def get_main_activity(self):
		package_name = self.get_package_name()
		for application in self.root.iter('application'):
			for activity in application.iter('activity'):
				activity_name = activity.get(AndroidManifest.ANDROID_NAMESPACE + 'name')
				# add package name
				if activity_name[0] == '.':
					activity_name = package_name + activity_name
				# find mainActivity
				for inter in activity.iter('intent-filter'):
					for action in inter.iter('action'):
						if action.get(AndroidManifest.ANDROID_NAMESPACE +
						              'name') == 'android.intent.action.MAIN':
							for cate in inter.iter('category'):
								if cate.get(AndroidManifest.ANDROID_NAMESPACE +
								            'name') == 'android.intent.category.LAUNCHER':
									return activity_name
		return None



import json
import shutil
import sys
from http.server import BaseHTTPRequestHandler, HTTPServer

from colorama import Back, Fore, Style, init


def get_output_server_file_path():
	return __file__


color_idx = 0
colors = [Fore.RED, Fore.GREEN, Fore.YELLOW, Fore.BLUE, Fore.MAGENTA, Fore.CYAN]
id_to_color = dict()


def _inc_color():
	global color_idx
	color_idx = (color_idx + 1) % len(colors)


def _call_num_exists(call_num):
	global id_to_color
	return call_num in id_to_color.keys()


def _get_color(call_num):
	global color_idx
	global colors
	global id_to_color

	if not _call_num_exists(call_num):
		# enter
		color = colors[color_idx]
		id_to_color[call_num] = color
		_inc_color()
	else:
		# leave
		color = id_to_color[call_num]
		del id_to_color[call_num]

	return color


def get_terminal_width():
	width, _ = shutil.get_terminal_size(fallback=(80, 24))

	# The Windows get_terminal_size may be bogus, let's sanify a bit.
	if width < 40:
		width = 80

	return width


def print_title(title, color):
	print(color + entire_row_str(title))


def print_data(data):
	print(json.dumps(data, indent=4, sort_keys=True))


def entire_row_str(title=None, sepchar='='):
	fullwidth = get_terminal_width()

	# The goal is to have the line be as long as possible
	# under the condition that len(line) <= fullwidth.
	if sys.platform == "win32":
		# If we print in the last column on windows we are on a
		# new line but there is no way to verify/neutralize this
		# (we may not know the exact line width).
		# So let's be defensive to avoid empty lines in the output.
		fullwidth -= 1
	if title is not None:
		# we want 2 + 2*len(fill) + len(title) <= fullwidth
		# i.e.    2 + 2*len(sepchar)*N + len(title) <= fullwidth
		#         2*len(sepchar)*N <= fullwidth - len(title) - 2
		#         N <= (fullwidth - len(title) - 2) // (2*len(sepchar))
		N = max((fullwidth - len(title) - 2) // (2 * len(sepchar)), 1)
		fill = sepchar * N
		line = f"{fill} {title} {fill}"
	else:
		# we want len(sepchar)*N <= fullwidth
		# i.e.    N <= fullwidth // len(sepchar)
		line = sepchar * (fullwidth // len(sepchar))
	# In some situations there is room for an extra sepchar at the right,
	# in particular if we consider that with a sepchar like "_ " the
	# trailing space is not important at the end of the line.
	if len(line) + len(sepchar.rstrip()) <= fullwidth:
		line += sepchar.rstrip()

	return line


def print_hook_details(data):
	call_num = data['callNum']
	method_name = data['methodName']
	class_name = data['className']
	param_types = data['paramTypes']
	params = data['params']

	if not _call_num_exists(call_num):
		enter = True
		data_to_print = {
			'Param Types': param_types,
			'Params': params
		}
	else:
		enter = False
		data_to_print = {
			'Return Value': data['retval']
		}

	color = _get_color(call_num)

	title = ('[ENTER]' if enter else '[LEAVE]') + ' ' + class_name + "::" + method_name

	print_title(title, color)
	print_data(data_to_print)


def pprint(msg):
	if type(msg) == dict:
		msg_type = msg['type']

		if msg_type == 'hook-enter' or msg_type == 'hook-leave':
			data = msg['data']
			print_hook_details(data)

		elif msg_type == "error":
			data = msg['data']
			print(
				Back.RED + Fore.WHITE + Style.BRIGHT + "[ERROR]" + Style.RESET_ALL + '\t' + Fore.RED + Style.BRIGHT + data)
	else:
		print(msg)

	print()


class OutputWindowServer(BaseHTTPRequestHandler):

	def log_message(self, format, *args):
		return

	def do_POST(self):
		self.send_response(200)
		self.end_headers()
		length = int(self.headers['Content-length'])
		content = self.rfile.read(length)
		msg = json.loads(content)['msg']

		if msg == FRIDA_SERVER_KILL_COMMAND:
			exit()

		pprint(msg)


if __name__ == "__main__":
	from consts import FRIDA_SERVER_KILL_COMMAND

	init(autoreset=True)

	if len(sys.argv) != 4:
		exit(-1)

	_, device_id, app_name, port = sys.argv


	def aux_print(name, value):
		pad = 15
		print(Fore.GREEN + Style.BRIGHT + name.ljust(pad) + Style.RESET_ALL + str(value))


	aux_print("Device", device_id)
	aux_print("Package", app_name)
	print()

	server_address = ('', int(port))
	httpd = HTTPServer(server_address, OutputWindowServer)
	httpd.allow_reuse_address = True
	httpd.serve_forever()



"""
A wrapper class - represents a Java instance.
"""


class Instance:

	def __init__(self, instance_data, agent):
		self.class_name = instance_data.get('class_name')
		self.id = instance_data['id']

		self.agent = agent

	def __repr__(self):
		class_str = "class='{0}', ".format(self.class_name) if self.class_name else ''
		return "Instance({0}toString='{1}')".format(class_str, self.__str__())

	def __str__(self):
		return self.agent.__to_string__(self.id)

	def get_fields(self, name_substr=None):
		return self.agent.get_fields(self.class_name, is_static=False, name_substr=name_substr)

	def get(self, field_name):
		return self.agent.__get_field_value__(self.class_name, field_name, instance_id=self.id)

	def call(self, method_name, params=[], param_types=[]):
		return self.agent.__call_instance_method__(self.id, method_name, params, param_types)

	def to_string(self):
		return self.agent.__to_string__(self.id)



from colorama import Back, Fore, Style


class Log:
	DEBUG = 'DEBUG'
	INFO = 'INFO'
	WARN = 'WARN'
	ERROR = 'ERROR'

	levels_mapping = {
		DEBUG: 0,
		INFO: 1,
		WARN: 2,
		ERROR: 3,
	}

	min_level = levels_mapping[INFO]

	@staticmethod
	def check_level(level):
		return Log.levels_mapping[level] >= Log.min_level

	@staticmethod
	def set_log_level(level):
		Log.min_level = Log.levels_mapping[level]

	@staticmethod
	def debug(s):
		if Log.check_level(Log.DEBUG):
			print(Back.BLUE + Fore.WHITE + Style.BRIGHT + "[DEBUG]" +
			      Style.RESET_ALL + '\t' + Fore.BLUE + Style.BRIGHT + s +
			      Style.RESET_ALL)

	@staticmethod
	def info(s):
		if Log.check_level(Log.INFO):
			print(Back.BLACK + Fore.WHITE + "[INFO]" + Style.RESET_ALL + '\t' +
			      Fore.WHITE + Style.BRIGHT + s + Style.RESET_ALL)

	@staticmethod
	def warn(s):
		if Log.check_level(Log.WARN):
			print(Back.YELLOW + Fore.WHITE + Style.BRIGHT + "[WARNING]" +
			      Style.RESET_ALL + '\t' + Fore.YELLOW + Style.BRIGHT + s +
			      Style.RESET_ALL)

	@staticmethod
	def error(s):
		if Log.check_level(Log.ERROR):
			print(Back.RED + Fore.WHITE + Style.BRIGHT + "[ERROR]" + Style.RESET_ALL +
			      '\t' + Fore.RED + Style.BRIGHT + s + Style.RESET_ALL)


def debug(s):
	log(s, Log.DEBUG)


def info(s):
	log(s, Log.INFO)


def warn(s):
	log(s, Log.WARN)


def error(s):
	log(s, Log.ERROR)


def log(s, level=Log.INFO, do_style=True):
	if not do_style:
		print(s)
		return

	getattr(Log, level.lower())(s)



"""
A wrapper class - represents a Java field.
"""


class Field:

	def __init__(self, field_data, agent):
		self.class_name = field_data['class_name']
		self.name = field_data['name']
		self.declaration = field_data['declaration']
		self.type = field_data['type']
		self.is_static = field_data['is_static']

		self.agent = agent

	def __repr__(self):
		return "[{static}Field(Name={name}, Type={type})]".format(
			name=self.name,
			type=self.type,
			static="Static " if self.is_static else '')

	def __str__(self):
		return self.__repr__()



"""
A wrapper class - represents a Java method.
"""


class Method:

	def __init__(self, method_data, agent):
		self.class_name = method_data['class_name']
		self.name = method_data['name']
		self.signature = method_data['signature']
		self.is_static = method_data['is_static']
		self.return_type = method_data['return_type']
		self.param_types = method_data['param_types']

		self._agent = agent

	def __repr__(self):
		params_str = '' if self.param_types == [] else ', ParamTypes={param_types}'.format(param_types=self.param_types)
		return "[{static}Method(Name={name}, ReturnType={return_type}{params_str})]".format(
			name=self.name,
			static="Static " if self.is_static else '',
			return_type=self.return_type,
			params_str=params_str)

	def __str__(self):
		return self.__repr__()



"""
A wrapper class - represents a Java array.
"""

from .winstance import Instance
from ...utils import simplify_java_array_type, extract_elem_type_from_simple_arr_type


class ArrayIterator:
	def __init__(self, arr):
		self._arr = arr
		self._index = 0

	def __next__(self):
		if self._index < len(self._arr):
			res = self._arr[self._index]
			self._index += 1
			return res

		raise StopIteration


class Array(Instance):

	def __init__(self, array_data, agent):
		array_data['class_name'] = simplify_java_array_type(array_data.get('class_name'))

		super().__init__(array_data, agent)

		self.elem_type = None
		if self.class_name:
			self.elem_type = extract_elem_type_from_simple_arr_type(self.class_name)

		# calculate length of array
		self.size = self.length()

	def __iter__(self):
		return ArrayIterator(self)

	def __repr__(self):
		return self.__str__()

	def __str__(self):
		s = '['
		for i in range(self.size):
			item = self[i]
			s += str(item)
			if i < self.size - 1:
				s += ', '
		s += ']'
		return s

	def _validate_index(self, index):
		if self.size == 0:
			raise IndexError("Array size is 0")
		if index < 0 or index >= self.size:
			raise IndexError("Index is out of range [0, {0}]".format(self.size - 1))

	def __getitem__(self, index):
		self._validate_index(index)

		value_data = self.agent.api.array_get_item(self.id, index)
		return self.agent._value_data_to_python_value(value_data)

	def __setitem__(self, index, value):
		self._validate_index(index)

		value_to_set = self.agent._convert_to_api_js_value(value)

		self.agent.api.array_set_item(self.id, index, value_to_set)

	def __len__(self):
		return self.size

	def length(self):
		return self.agent.api.get_array_length(self.id)

	def to_list(self):
		return [x for x in self]



from .log import debug
from .utils import *


class FilesOperations:

	def __init__(self, device):
		self._device = device

	def pull(self, remote_path, local_path):
		debug("Pulling file from device '{0}': '{1}' -> '{2}'".format(
			self._device.id, remote_path, local_path))
		check_output("{0} pull {1} {2}".format(self._device.__adb_prefix__(),
		                                       remote_path, local_path))

	def push(self, local_path, remote_path):
		debug("Puhsing file to device '{0}': '{1}' -> '{2}'".format(
			self._device.id, local_path, remote_path))
		check_output("{0} push {1} {2}".format(self._device.__adb_prefix__(),
		                                       local_path, remote_path))

	def exists(self, remote_path):
		debug("Checking if file exists on device '{0}': '{1}'".format(
			self._device.id, remote_path))
		return self._device.shell(
			'ls {0} > /dev/null 2>&1; echo $?'.format(remote_path)) == '0'

	def remove(self, remote_path):
		if not self.exists(remote_path):
			debug("Remote path does not exist ('{0}')".format(remote_path))
			return False

		debug("Removing '{0}' from device '{1}'".format(
			remote_path, self._device.id))
		return self._device.shell(
			'rm -rf {0} > /dev/null 2>&1; echo $?'.format(remote_path)) == '0'



"""
A Pythonic API (or wrapper) for `frida`'s-like operations on an Android app.
This object is attached to a specific app on a specific device.
It lets you hook and call methods, change field's value, and more.
Can be retrieved by calling :py:meth:`~marrs.app.App.attach_frida_agent` method on an :py:class:`~marrs.app.App` object.
"""

import json
import os

import frida

from .hooks.hooks_manager import HooksManager
from .output_window_client import OutputWindowClient
from .utils import is_simple_array_type, \
	convert_to_api_class_name, java_primitive_type_to_wrapper_class, is_java_class_type, \
	extract_base_elem_type_from_simple_arr_type
from .wrappers import Field, Method, Constructor, Class
from .wrappers.instance_factory import InstanceFactory
from .wrappers.instances import Instance
from ..utils import get_dir_path, read_file


class FridaAgent:

	def __init__(self, app, initial_script=None):
		self.app = app
		self.output = OutputWindowClient(device_id=app.device.id, app_name=app.name)
		self.hooks = HooksManager(self)

		self.app.device.__log__(
			"Creating new frida agent for app", extra_info=app.name)

		if initial_script:
			self.app.device.__warn__(
				"	Initial script is not empty and app is running so restarting app")
			app.force_stop()
			device = frida.get_device(self.app.device.id, timeout=1)
			self.app.device.__log__("	Spawning app", extra_info=self.app.name)
			pid = device.spawn([self.app.name])
			self.app.device.__log__(
				"	Attaching to app", extra_info="pid={pid}".format(pid=pid))
			self.session = device.attach(pid)

			self.__load_api__()
			self.load_script(initial_script)

			device.resume(pid)
			return

		if not app.is_running():
			self.app.device.__log__("	App is not running, starting app")
			app.start()

		self.session = frida.get_device(
			self.app.device.id, timeout=1).attach(self.app.name)
		self.__load_api__()

	def __del__(self):
		self.kill()

	def __print__(self, s):
		if self.output.started:
			self.output.print(s)
		else:
			print(s)

	def __load_api__(self):
		api_js_path = os.path.join(get_dir_path(__file__), '..', 'data', 'bundle.js')
		self.api_script = self.load_script_from_file(api_js_path)
		self.api = self.api_script.exports

	def __on_message__(self, msg, data):
		# print("__on_message__", msg, data)
		msg_type = msg['type']

		if msg_type == 'send':
			payload = json.loads(msg['payload'])
			payload_type = payload['type']
			if payload_type == 'hook-enter':
				hook_id = payload['hookId']
				hook = self.hooks[hook_id]
				params = payload['params']
				orig_retval = payload['originalRetVal']
				call_num = hook.inc_call_num()

				if hook.show_output:
					self.__print__({
						'type': payload_type,
						'data': {
							'params': params,
							'methodName': hook.method_name,
							'className': hook.class_name,
							'paramTypes': hook.param_types,
							'callNum': call_num
						},
					})

				retval = self.hooks.call(hook_id, params, orig_retval)

				if hook.show_output:
					self.__print__({
						'type': 'hook-leave',
						'data': {
							'params': params,
							'methodName': hook.method_name,
							'className': hook.class_name,
							'paramTypes': hook.param_types,
							'callNum': call_num,
							'retval': retval
						},
					})

				self.api_script.post({
					'type': 'continue@' + str(hook_id),
					'payload': {
						'retval': self._convert_to_api_js_value(retval)
					}
				})

		elif msg_type == 'error':
			self.__print__({
				'type': 'error',
				'data': msg['stack']
			})

	def _get_params_data(self, params, param_types):
		converted_params = [self._convert_to_api_js_value(params[i], param_types[i] if param_types else None) for i in
		                    range(len(params))]

		if param_types is not None:
			param_types = [convert_to_api_class_name(pt) for pt in param_types]

		params_data = json.dumps({
			'params': converted_params,
			'paramTypes': param_types
		})

		return params_data

	def _value_data_to_python_value(self, value_data):
		# check if simple value
		if not isinstance(value_data, dict):
			return value_data

		return InstanceFactory.create(self, value_data)

	def _convert_to_api_js_value(self, python_value, elem_type=None):
		if python_value is None:
			return None

		if isinstance(python_value, list):
			return [self._convert_to_api_js_value(x, elem_type) for x in python_value]

		# If type is class type and value is not an instance, try to create new instance of that type
		if is_java_class_type(elem_type) and not isinstance(python_value, Instance):
			python_value = self.new_instance(elem_type, [str(python_value)])

		if isinstance(python_value, Instance):
			return json.dumps({
				'id': python_value.id,
			})

		return json.dumps({
			'value': python_value
		})

	def load_script_from_file(self, file_path):
		"""
		Load and run a script from file.

		:param file_path: The script's file path
		:type file_path: str
		:return: `frida`'s script object
		"""
		script_code = read_file(file_path)
		return self.load_script(script_code)

	def load_script(self, script_code):
		"""
		Load and run a script.

		:param script_code: The script's code to run
		:type script_code: str
		:return: `frida`'s script object
		"""
		script = self.session.create_script(script_code)
		script.on('message', self.__on_message__)
		script.load()
		return script

	def kill(self):
		"""
		Kills the agent's output window (if exist) and detached from app.
		"""

		# TODO: Unloading all the scripts ?
		self.output.kill()
		self.app.frida_agent = None

	# API methods
	def get_loaded_classes(self, regex_pattern=None):
		"""
		Get a list of loaded classes.

		:param regex_pattern: Match classes' names to this regex pattern
		:type regex_pattern: str
		:return: A list of loaded classes
		:rtype: A list of :py:class:`~marrs.frida.wrappers.wclass.Class`
		"""
		classes = self.api.get_loaded_classes(regex_pattern)
		classes = list(map(lambda c: Class(c, self), classes))
		return classes

	def get_methods(self, class_name, is_static=None, name_substr=None):
		"""
		Get a list of methods of a specific class.

		:param class_name: The full name of the class to get its methods
		:type class_name: str
		:param is_static: Get only static methods. None value (the default) will return both.
		:type is_static: bool
		:param name_substr: Match methods' names that contain this substring
		:type name_substr: str
		:return: A list of methods
		:rtype: A list of :py:class:`~marrs.frida.wrappers.wmethod.Method`
		"""
		class_name = convert_to_api_class_name(class_name)
		methods = self.api.get_methods_of_class(class_name)
		methods = list(map(lambda m: Method(m, self), methods))

		if is_static is not None:
			methods = list(filter(lambda m: m.is_static == is_static, methods))

		if name_substr is not None:
			methods = list(filter(lambda m: name_substr in m.name, methods))

		return methods

	def get_fields(self, class_name, is_static=None, name_substr=None):
		"""
		Get a list of fields of a specific class.

		:param class_name: The full name of the class to get its fields
		:type class_name: str
		:param is_static: Get only static fields. None value (the default) will return both.
		:type is_static: bool
		:param name_substr: Match fields' names that contain this substring
		:type name_substr: str
		:return: A list of fields
		:rtype: A list of :py:class:`~marrs.frida.wrappers.wfield.Field`
		"""
		if class_name is None:
			raise ValueError('class_name is None')

		class_name = convert_to_api_class_name(class_name)
		fields = self.api.get_fields_of_class(class_name)

		fields = list(map(lambda f: Field(f, self), fields))

		if is_static is not None:
			fields = list(filter(lambda f: f.is_static == is_static, fields))

		if name_substr is not None:
			fields = list(filter(lambda f: name_substr in f.name, fields))

		return fields

	def get_instances(self, class_name):
		"""
		Get all the living instances of a given class.

		:param class_name: The full name of the class to get its living instances
		:type class_name: str
		:return: A list of instances
		:rtype: A list of :py:class:`~marrs.frida.wrappers.winstance.Instance`
		"""
		class_name = convert_to_api_class_name(class_name)
		instances_data = self.api.get_instances_of_class(class_name)
		instances_data = list(map(lambda idata: InstanceFactory.create(self, idata), instances_data))
		return instances_data

	def get_class(self, class_name):
		"""
		Get a loaded class by its name.

		:param class_name: The full name of the class
		:type class_name: str
		:return: A class object
		:rtype: :py:class:`~marrs.frida.wrappers.wclass.Class`
		"""
		class_name = convert_to_api_class_name(class_name)
		class_data = self.api.get_class(class_name)
		return Class(class_data, self)

	def get_constructors(self, class_name):
		"""
		Get a list of constructors of a given class.

		:param class_name: The full name of the class
		:type class_name: str
		:return: A list of constructors objects
		:rtype: A list of :py:class:`~marrs.frida.wrappers.wconstructor.Constructor`
		"""
		constructors = self.api.get_constructors_of_class(class_name)
		constructors = list(map(lambda c: Constructor(c, self), constructors))
		return constructors

	def new_instance(self, class_name, params=[], param_types=None):
		"""
		Create a new instance of a given class.

		Can be used as method's param or a field's value.

		Examples:

		.. code-block:: python

		   agent.new_instance("int", [3])
		   agent.new_instance("double", [3.3])
		   agent.new_instance("java.lang.String", ['somestr'])
		   agent.new_instance("com.example.SomeClass", [[[1, 2], [3, 2]]], ['int[][]'])
		   agent.new_instance("com.example.SomeClass", [1, 1, '1', '1', False, 1, 1, 1],
		                                    ['java.lang.Integer', 'java.lang.Double', 'java.lang.Character',
		                                    'java.lang.String', 'java.lang.Boolean', 'java.lang.Byte',
		                                    'java.lang.Long', 'java.lang.Float'])
		   agent.new_instance("com.example.SomeClass", [agent.new_instance('java.lang.Integer', [3])])

		:param class_name: The full name of the class to create instance of
		:type class_name: str
		:param params: The constructor's parameters list
		:type params: list
		:param param_types: The constructor's parameters types list. This param is optional, if the constructor is overloaded you might need to specify the constructor param types.
		:type param_types: list
		:return: The new created instance
		:rtype: :py:class:`~marrs.frida.wrappers.winstance.Instance`
		"""
		if is_simple_array_type(class_name):
			raise ValueError("Use '{0}' function to create an array".format(self.new_array.__name__))

		if param_types is not None and len(params) != len(param_types):
			raise ValueError(
				"len(params) != len(param_types), className={0}, params={1}, param_types={2}".format(class_name, params,
				                                                                                     param_types))

		class_name = java_primitive_type_to_wrapper_class(class_name)

		params_data = self._get_params_data(params, param_types)
		instance_data = self.api.create_instance(class_name, params_data)
		return InstanceFactory.create(self, instance_data)

	def new_array(self, arr_type, arr_list):
		"""
		Create a new array instance.

		Can be used as method's param or a field's value.

		Examples:

		.. code-block:: python

		   agent.new_array("int[]", [1, 2, 3])
		   agent.new_array("int[][]", [[18, 2, 3], [12, 1, 1], [2, 2, 2]])
		   agent.new_array('java.lang.Integer[][][]', [[[1]], [[2], [9]], [[3]]])
		   agent.new_array("java.lang.Integer[]", [agent.new_instance('java.lang.Integer', [1]),
	                                               agent.new_instance('java.lang.Integer', [2]),
	                                               agent.new_instance('java.lang.Integer', [3])])


		:param arr_type: The type of the array.
		:type arr_type: str
		:param arr_list: The items of the array
		:type arr_list: list
		:return: The new created array
		:rtype: :py:class:`~marrs.frida.wrappers.warray.Array`
		"""
		if not is_simple_array_type(arr_type):
			raise ValueError("arr_type is not an array type")

		if not isinstance(arr_list, list):
			raise ValueError("The given arr_list must be of a list type")

		elem_type = extract_base_elem_type_from_simple_arr_type(arr_type)
		converted_elems = [self._convert_to_api_js_value(arr_list[i], elem_type) for i in range(len(arr_list))]

		arr_type = convert_to_api_class_name(arr_type)

		array_data = self.api.create_array(arr_type, converted_elems)
		return InstanceFactory.create(self, array_data)

	def __to_string__(self, instance_id):
		return self.api.to_string(instance_id)

	def __call_instance_method__(self, instance_id, method_name, params=[], param_types=[]):
		params_data = self._get_params_data(params, param_types)
		value_data = self.api.call_method(method_name, params_data, instance_id, None)
		return self._value_data_to_python_value(value_data)

	def call_static_method(self, class_name, method_name, params=[], param_types=[]):
		"""
		Call a static method of a class (and get the result value).

		Examples:

		.. code-block:: python

		   ret_val = agent.call_static_method("com.example.SomeClass", "someStaticMethod1", ["someStr"], ["java.lang.String"])
		   ret_val = agent.call_static_method("com.example.SomeClass", "someStaticMethod2", [1, 3], ['int', 'double'])
		   ret_val = agent.call_static_method("com.example.SomeClass", "someNotOverloadedMethod", ["someStr", 3])
		   ret_val = agent.call_static_method("com.example.SomeClass", "someNoParamsMethod")


		:param class_name: The name of the class
		:type class_name: str
		:param method_name: The name of the method to call to
		:type method_name: str
		:return: The return value of the method
		"""
		class_name = convert_to_api_class_name(class_name)
		params_data = self._get_params_data(params, param_types)
		value_data = self.api.call_method(method_name, params_data, None, class_name)
		return self._value_data_to_python_value(value_data)

	def get_static_field_value(self, class_name, field_name):
		"""
		Get the value of static field of a class.

		For example:

		.. code-block:: python

		   ret_val = agent.get_static_field_value("com.example.SomeClass", "someStaticFieldName")


		:param class_name: The name of the class
		:type class_name: str
		:param field_name: The name of the field to get its value
		:type field_name: str
		:return: The value of the field
		"""
		return self.__get_field_value__(class_name, field_name)

	def __get_field_value__(self, class_name, field_name, instance_id=None):
		class_name = convert_to_api_class_name(class_name)
		value_data = self.api.get_field_value(class_name, field_name, instance_id)
		return self._value_data_to_python_value(value_data)

	def set_static_field_value(self, class_name, field_name, new_value):
		"""
		Set the value of static field of a class.

		For example:

		.. code-block:: python

		   ret_val = agent.get_static_field_value("com.example.SomeClass", "someStaticFieldName")


		:param class_name: The name of the class
		:type class_name: str
		:param field_name: The name of the field to get set its value
		:type field_name: str
		:param new_value: The new value to set
		"""
		self.__set_field_value__(class_name, field_name, new_value)

	def __set_field_value__(self, class_name, field_name, new_value, instance_id=None):
		class_name = convert_to_api_class_name(class_name)
		value_to_set = self._convert_to_api_js_value(new_value)
		self.api.set_field_value(class_name, field_name, value_to_set, instance_id)

	def run_js(self, js_code):
		"""
		Run Javascript code and get the result.

		For example:

		.. code-block:: python

		   ret_val = agent.run_js("3+5")
		   assert(ret_val == 8)


		:param js_code: The Javascript code to run
		:type js_code: str
		:return: The result value
		"""
		value_data = self.api.run_js(js_code)
		return self._value_data_to_python_value(value_data)



import os
from os import path
from pathlib import Path

import appdirs

from .consts import PACKAGE_NAME, STORAGE_TMP_DIR_NAME
from .log import debug
from .utils import copy_file_with_dir_creation, get_filename, delete_folder, get_temp_name


class Storage:
	STORAGE_DIR_PATH = str(Path(appdirs.user_data_dir(appname=PACKAGE_NAME)))

	@staticmethod
	def get_folder_path():
		return Storage.STORAGE_DIR_PATH

	@staticmethod
	def set_folder_path(storage_dir_path):
		Storage.STORAGE_DIR_PATH = storage_dir_path

	@staticmethod
	def _get_path(filename):
		return path.join(Storage.STORAGE_DIR_PATH, filename)

	@staticmethod
	def contains(filename):
		return Storage.get(filename) is not None

	@staticmethod
	def get(filename):
		filepath = Storage._get_path(filename)
		if path.exists(filepath):
			return filepath

		return None

	@staticmethod
	def delete():
		delete_folder(Storage.STORAGE_DIR_PATH)

	@staticmethod
	def get_temp_dir_path():
		return path.join(Storage.STORAGE_DIR_PATH, STORAGE_TMP_DIR_NAME)

	@staticmethod
	def delete_temp_dir():
		delete_folder(Storage.get_temp_dir_path())

	@staticmethod
	def create_temp_dir():
		tmp_dir_path = path.join(Storage.get_temp_dir_path(), get_temp_name())
		os.makedirs(tmp_dir_path)
		return tmp_dir_path

	@staticmethod
	def save(src_file_path, filename=None):
		if not filename:
			filename = get_filename(src_file_path)

		dest_file_path = Storage._get_path(filename)

		debug("Saving file to local storage: {0}".format(dest_file_path))
		copy_file_with_dir_creation(src_file_path, dest_file_path)



PACKAGE_NAME = 'marrs'

FRIDA_SERVER_VERSION = '14.2.18'
FRIDA_SERVER_FILENAME = 'frida-server-{version}-android-{arch}'
FRIDA_SERVER_DOWNLOAD_URL = 'https://github.com/frida/frida/releases/download/{version}/{filename}.xz'
FRIDA_SERVER_REMOTE_PATH = '/data/local/tmp/frida-server-{version}-{arch}'

STORAGE_TMP_DIR_NAME = 'tmp'



class Hook:
	def __init__(self, hooks_manager, hook_id, class_name, method_name, param_types, hook_impl, show_output):
		self._hooks_manager = hooks_manager
		self.id = hook_id
		self.class_name = class_name
		self.method_name = method_name
		self.param_types = param_types
		self.hook_impl = hook_impl
		self.enabled = True
		self.show_output = show_output
		self.call_num = 0

	def __repr__(self):
		return "Hook(id='{0}', class='{1}', method='{2}', param_types='{3}', enabled={4}, show_output={5})".format(
			self.id,
			self.class_name,
			self.method_name,
			self.param_types,
			self.enabled,
			self.show_output)

	def enable(self):
		self._hooks_manager.enable(self.id)

	def disable(self):
		self._hooks_manager.disable(self.id)

	def inc_call_num(self):
		self.call_num += 1
		return self.call_num



"""
A wrapper class - represents a Java class.
"""


class Class:

	def __init__(self, class_data, agent):
		self.agent = agent

		self.name = class_data['name']

	def __repr__(self):
		return self.__str__()

	def __str__(self):
		return "Class('{0}')".format(self.name)

	def get_methods(self, is_static=None, name_substr=None):
		return self.agent.get_methods(self.name, is_static=is_static, name_substr=name_substr)

	def get_fields(self, is_static=None, name_substr=None):
		return self.agent.get_fields(self.name, is_static=is_static, name_substr=name_substr)

	def get_constructors(self):
		return self.agent.get_constructors(self.name)

	def get_instances(self):
		return self.agent.get_instances(self.name)

	def set_field(self, field_name, value):
		self.agent.set_static_field_value(self.name, field_name, value)

	def get_field(self, field_name):
		return self.agent.get_static_field_value(self.name, field_name)

	def new(self, params=[], param_types=None):
		return self.agent.new_instance(self.name, params, param_types)



import json
import platform
import random
import sys
from os import system

import requests

from .consts import FRIDA_SERVER_KILL_COMMAND
from .output_window_server import get_output_server_file_path
from ..log import debug


class OutputWindowClient:

	def __init__(self, device_id, app_name):
		self.port = random.randint(30000, 60000)
		self.device_id = device_id
		self.app_name = app_name
		self.started = False

	def start(self):
		file_path = get_output_server_file_path()

		plat = platform.system()
		if plat == 'Windows':
			cmd = 'start cmd /c "\"{0}\" \"{1}\" {2} {3} {4}"'.format(
				sys.executable, file_path, self.device_id, self.app_name, self.port)
		else:
			raise Exception("'{0}' platform is not supported yet".format(plat))

		debug(cmd)
		res = system(cmd)

		self.started = True

	def kill(self):
		if self.started:
			self.print(FRIDA_SERVER_KILL_COMMAND)

	def print(self, s):
		if not self.started:
			return

		requests.post(
			'http://localhost:{0}'.format(self.port), data=json.dumps({'msg': s}))



"""
Represents an Android Application.

This object can be retrieved by calling :py:meth:`~marrs.device.Device.get_app`,
:py:meth:`~marrs.device.Device.get_apps` or :py:meth:`~marrs.device.Device.install_app` methods on a :py:class:`~marrs.device.Device` object.
"""

from .android_manifest import AndroidManifest
from .frida import FridaAgent
from .storage import Storage
from .utils import *


class App:

	def __init__(self, device, package_name, remote_apk_path):
		self.device = device
		self.name = package_name
		self.remote_apk_path = remote_apk_path
		self.local_apk_path = None
		self.main_activity = None
		self.frida_agent = None
		self._tmp_dir = None

	def __str__(self):
		return self.name

	def __repr__(self):
		return "App('{0}')".format(self.name)

	def __del__(self):
		if self._tmp_dir:
			delete_folder(self._tmp_dir)

	# if self.frida_agent:
	#	self.frida_agent.kill()

	def is_installed(self):
		"""
		Check wether this app is installed on the device.

		:return: True if app is installed on the device, False otherwise
		:rtype: bool
		"""
		return self.device.is_app_installed(self.name)

	def install(self):
		"""
		Install this app on the device.
		The app's APK file is taken from the device.
		"""
		self.device.install_app(self.get_local_apk_path())

	def uninstall(self):
		"""
		Uninstall the application from the device.
		"""
		self.device.uninstall_app(self.name)

	def is_running(self):
		"""
		Check if this app is currently running on the device.

		:return: True if app is currently running, False otherwise.
		:rtype: bool
		"""
		return self.device.is_app_running(self.name)

	def clear_data(self):
		"""
		Clear application data.
		"""
		self.device.clear_app_data(self.name)

	def force_stop(self):
		"""
		Stop the application from running.
		"""
		self.device.force_stop_app(self.name)

	def start(self):
		"""
		Start the application.
		"""
		self.device.start_app(self.name, self.get_main_activity())

	def get_version(self):
		"""
		Get the version of the application.

		:return: The app's version if installed, None otherwise
		:rtype: str
		"""
		return self.device.get_app_version(self.name)

	def get_local_apk_path(self):
		"""
		Returns the local path to the app's APK file.
		If the APK doesn't exist on local machine, it will be pulled from the device to package's local storage.

		:return: The local path to the app's APK file
		:rtype: str
		"""
		if self.local_apk_path:
			return self.local_apk_path

		self._tmp_dir = Storage.create_temp_dir()
		local_apk_path = path.join(self._tmp_dir, get_filename(self.remote_apk_path, with_extension=True))
		self.device.files.pull(self.remote_apk_path, local_apk_path)
		self.local_apk_path = local_apk_path
		return self.local_apk_path

	def get_main_activity(self):
		"""
		Returns the name of the main activity of the app.

		:return: The name of the main activity.
		:rtype: str
		"""
		if self.main_activity:
			return self.main_activity

		local_apk_path = self.get_local_apk_path()
		manifest = AndroidManifest.from_apk(local_apk_path)
		self.main_activity = manifest.get_main_activity()
		return self.main_activity

	def attach_frida_agent(self, initial_script=None):
		"""
		Attach a :py:class:`~marrs.frida.agent.FridaAgent` to the app.

		:param initial_script: An initial script to run while the app is spawned.
		:type initial_script: str
		:return: An :py:class:`~marrs.frida.agent.FridaAgent` object or None if an error occured
		:rtype: :py:class:`~marrs.frida.agent.FridaAgent`
		"""
		self.device.__log__("Attaching frida agent to app", extra_info=self.name)

		if self.frida_agent is not None:
			self.device.__error__("	Frida agent is already attached to this app")
			return None

		if not self.device.is_rooted():
			self.device.__error__("	Device must be rooted in order to use frida")
			return None

		if not self.device.is_frida_server_running():
			self.device.__warn__("	Frida server is not running on the device")
			self.device.run_frida_server()

		self.frida_agent = FridaAgent(self, initial_script)
		return self.frida_agent



import re

from ..utils import invert_dict

JAVA_PRIMITIVE_TYPE_TO_WRAPPER_CLASS_MAPPING = {
	'int': 'java.lang.Integer',
	'boolean': 'java.lang.Boolean',
	'char': 'java.lang.Character',
	'float': 'java.lang.Float',
	'double': 'java.lang.Double',
	'byte': 'java.lang.Byte',
	'short': 'java.lang.Short',
	'long': 'java.lang.Long',
}

JAVA_PRIMITIVE_TYPES = JAVA_PRIMITIVE_TYPE_TO_WRAPPER_CLASS_MAPPING.keys()

JAVA_PRIMITIVE_TYPE_TO_ARRAY_LETTER_MAPPING = {
	'byte': 'B',
	'short': 'S',
	'int': 'I',
	'long': 'J',
	'float': 'F',
	'double': 'D',
	'boolean': 'Z',
	'char': 'C',
}

JAVA_CLASS_TYPE_ARRAY_LETTER = 'L'

ARRAY_LETTER_TO_JAVA_PRIMITIVE_TYPE_MAPPING = invert_dict(JAVA_PRIMITIVE_TYPE_TO_ARRAY_LETTER_MAPPING)


def is_java_primitive_type(class_name):
	return class_name in JAVA_PRIMITIVE_TYPES


def primitive_type_to_letter(class_name):
	return JAVA_PRIMITIVE_TYPE_TO_ARRAY_LETTER_MAPPING.get(class_name, class_name)


def is_simple_array_type(class_name):
	return class_name and class_name.endswith('[]')


def is_java_array_type(class_name):
	return class_name and class_name.startswith('[')


def java_primitive_type_to_wrapper_class(prim_type):
	return JAVA_PRIMITIVE_TYPE_TO_WRAPPER_CLASS_MAPPING.get(prim_type, prim_type)


def is_java_class_type(class_name):
	return class_name is not None and not is_java_primitive_type(class_name) and not is_simple_array_type(
		class_name) and not is_java_array_type(
		class_name)


def convert_to_api_class_name(class_name):
	# for now only deal with converting simple array name
	if is_simple_array_type(class_name):
		groups = re.search(r"(.+?)((?:\[\])+)", class_name).groups()
		elem_type = groups[0]
		brackets = groups[1]

		prefix = '[' * (len(brackets) // 2)
		if is_java_primitive_type(elem_type):
			letter = primitive_type_to_letter(elem_type)
			return prefix + letter

		# class type
		return prefix + JAVA_CLASS_TYPE_ARRAY_LETTER + elem_type + ';'

	return class_name


def get_java_array_type(elem_type, depth):
	prefix = ('[' * (depth - 1))
	if is_java_primitive_type(elem_type):
		return prefix + JAVA_PRIMITIVE_TYPE_TO_ARRAY_LETTER_MAPPING[elem_type]

	return prefix + JAVA_CLASS_TYPE_ARRAY_LETTER + elem_type + ";"


def simplify_java_array_type(class_name):
	if not is_java_array_type(class_name):
		return class_name

	groups = re.search(r"(\[+)(\w)(.*)", class_name).groups()
	dimensions = len(groups[0])
	letter = groups[1]

	if letter == JAVA_CLASS_TYPE_ARRAY_LETTER:  # class type
		arr_type = groups[2][:-1]
	else:
		arr_type = ARRAY_LETTER_TO_JAVA_PRIMITIVE_TYPE_MAPPING[letter]

	return arr_type + ('[]' * dimensions)


def extract_elem_type_from_simple_arr_type(class_name):
	if not is_simple_array_type(class_name):
		return class_name

	return class_name[:-2]


def extract_base_elem_type_from_simple_arr_type(class_name):
	if not is_simple_array_type(class_name):
		return class_name

	groups = re.search(r"(.+?)((?:\[\])+)", class_name).groups()
	arr_type = groups[0]
	return arr_type



"""
A wrapper class - represents a Java class constructor.
"""


class Constructor:

	def __init__(self, constructor_data, agent):
		self.class_name = constructor_data['class_name']
		self.signature = constructor_data['signature']
		self.param_types = constructor_data['param_types']

		self.agent = agent

	def __repr__(self):
		params_str = '' if self.param_types == [] else ', ParamTypes={param_types}'.format(param_types=self.param_types)
		return "[Constructor(Class={class_name}{params_str})]".format(
			class_name=self.class_name,
			params_str=params_str)

	def __str__(self):
		return self.__repr__()

	def new(self, params=[]):
		return self.agent.new_instance(self.class_name, params, self.param_types)



from .device import Device
from .log import error
from .utils import check_output


def get_devices(rooted_only=False, adb_path='adb'):
	devices = check_output('"{0}" devices'.format(adb_path))

	devices = devices.strip().split('\n')[1:]  # skip first line

	def filter_func(line):
		parts = line.split()

		if len(parts) != 2:
			return False

		if parts[1] != 'device':
			return False

		return True

	def map_func(line):
		return Device(line.split()[0], adb_path)

	devices = list(map(map_func, filter(filter_func, devices)))

	if rooted_only:
		devices = list(filter(lambda d: d.is_rooted(), devices))

	return devices


def get_device(adb_path='adb'):
	"""
	Get the first device object.

	:param name: adb_path - The path to Android Debug Bridge binary
	:param type: str
	:return: Device
	"""
	devices = get_devices(adb_path=adb_path)

	if not devices:
		error("There are no connected devices")
		return None

	return devices[0]



from setuptools import setup, find_packages

setup(
	name='marrs',
	version='1.0.2',
	author='Oran Gilboa',
	author_email='oran.gilboa1@gmail.com',
	package_dir={'': 'src'},
	packages=find_packages(where='src'),
	package_data={'marrs': ['data/bundle.js']},
	scripts=[],
	url='https://github.com/oran1248/marrs',
	download_url='https://github.com/oran1248/marrs/archive/refs/tags/1.0.2.tar.gz',
	license='GNU GPL v3',
	description='Python package for Android Java apps researchers, built on top of tools like frida and adb',
	long_description='Marrs is a Python package for Android Java apps researchers, built on top of tools like frida and adb. Using Marrs you can write Python code that modifies fields\' value, calls methods, creates instances, hooks methods and more.',
	install_requires=[
		'frida~=14.2.18',
		'androguard~=3.3.5',
		'lxml~=4.6.2',
		'requests~=2.26.0',
		'colorama~=0.4.3',
		'appdirs~=1.4.4',
		'setuptools~=41.2.0'
	]
)
